SYSTEM_DATABASES = [
    "SNOWFLAKE",
    "WORKSHEETS_APP",
]

SYSTEM_SCHEMAS = [
    "PUBLIC",
    "INFORMATION_SCHEMA",
]

SYSTEM_ROLES = [
    "ACCOUNTADMIN",
    "ORGADMIN",
    "PUBLIC",
    "SECURITYADMIN",
    "SYSADMIN",
    "USERADMIN",
]

SYSTEM_USERS = [
    "SNOWFLAKE",
]

SYSTEM_SECURITY_INTEGRATIONS = [
    "APPLICA",
]
